;;; Author: David Lichteblau, Copyright 2003
;;; Parts taken from Douglas Crosher's eli-hacks.txt

(in-package :lep)
(declaim (optimize (debug 2)))

(define-condition lep-error (error) ())
(define-condition no-lep-connection (lep-error) ())
(define-condition lep-operation-aborted (lep-error) ())	;unused

(defvar *lep-is-running* nil)
(defvar *send-lock* (sb-thread:make-mutex :name "LEP stream lock"))
(defvar *lep-stream*)

(defvar *stream-protocol-lock* (sb-thread:make-mutex :name ":stream lock"))
(defvar *stream-protocol-requests* (make-hash-table))
(defvar *session-stream*)

(defvar *buffer-threads* (make-hash-table))

(defun lep-is-running ()
  *lep-is-running*)

(defun ensure-lep ()			;this should be called ASSERT-LEP
  (unless *lep-is-running*
    (error 'no-lep-connection)))

(defun invoke-with-stream (fn stream-or-pathname open-args)
  (if (streamp stream-or-pathname)
      (funcall fn stream-or-pathname)
      (with-open-stream (s (apply #'open stream-or-pathname open-args))
	(funcall fn s))))

(defmacro with-stream ((var stream-or-pathname &rest open-args) &body body)
  `(invoke-with-stream (lambda (,var) ,@body)
		       ,stream-or-pathname
		       (list ,@open-args)))

(define-condition user-error (simple-error) ())

(defun user-error (control &rest arguments)
  (error 'user-error :format-control control :format-arguments arguments))

(defun start-emacs-lisp-interface
    (&key default-port (announce-to-stream *standard-output*))
  (setf default-port (cond (default-port) (1025)))
  (flet ((doit ()
	   (let ((password (random (expt 2 24)))
		 (fd (make-instance 'sb-bsd-sockets:inet-socket
		       :type :stream
		       :protocol (sb-bsd-sockets:get-protocol-by-name "tcp"))))
	     (unwind-protect
		 (let ((port (bind-to-suitable-port fd default-port)))
		   (sb-bsd-sockets:socket-listen fd 5)
		   (with-stream (s announce-to-stream
				   :direction :output
				   :if-exists :supersede)
		     (format s "~D ~D :upper NIL 1" port password)
		     (finish-output s))
		   (setf *lep-is-running* t)
		   (accept-lep-connections fd password))
	       (setf *lep-is-running* nil)
	       (when fd
		 (sb-bsd-sockets:socket-close fd))))))
    (setf (gethash 1 *buffer-threads*) (sb-thread:current-thread-id))
    (make-thread "LEP connection listener" #'doit)))

(defun bind-to-suitable-port (fd default-port)
  (for ((port :from default-port))
    (handler-case
	(progn
	  (sb-bsd-sockets:socket-bind fd #(127 0 0 1) port)
	  (return port))
      (sb-bsd-sockets:address-in-use-error ()))))

(defun accept-lep-connections (listenfd password)
  (loop
    (multiple-value-bind (sockfd peer-address)
	(sb-bsd-sockets:socket-accept listenfd)
      (make-thread (format nil "Lisp session from ~A" peer-address)
		   'lep-login sockfd password))))

;;;; be careful (i.e., don't use READ) when parsing arguments from the
;;;; socket before the authentication phase has finished.  Even with
;;;; null *READ-EVAL*, bad stuff might happen.  Also, limit the number
;;;; of characters, so that we are not vulnerable to DoS attacks like an
;;;; idiot piping /dev/random to us.
;;;;
;;;; wishlist: unix domain socket support

(defun whitespacep (c)
  (member c '(#\space #\tab #\newline)))

(defvar *read-token-limit* 80)		;a line should really be enough

(defun read-token (stream)
  "Read a white-space delimited string from STREAM, or return NIL if the string
   would be unreasonably long.  Special handling of strings and backslashes."
  ;; skip whitespace
  (for ((i :from 0)
	(c = (peek-char nil stream nil nil))
	:while (whitespacep c))
    (when (eql i *read-token-limit*)
      (return-from read-token nil))
    (read-char stream))
  ;; read a token (whatever that is)
  (with-output-to-string (str)
    (let ((in-string-p nil)
	  (quotep nil))
      (for ((i :from 0)
	    (c = (read-char stream nil nil))
	    :while (and c (cond
			    (quotep t)
			    (in-string-p (not (char= c #\")))
			    (t (not (whitespacep c))))))
	(when (eql i *read-token-limit*)
	  (return-from read-token nil))
	(cond
	  (quotep
	    (princ c str)
	    (setf quotep nil))
	  (t
	    (case c
	      (#\" (setf in-string-p (not in-string-p)))
	      (#\\ (setf quotep t))
	      (t (princ c str)))))))))

;;;; read the handshake arguments and check the password.  Unauthorized
;;;; attempts should print a warning, but even here a limit is in order.
;;;; Else the terminal (or worse, the emacs buffer) might fill up with
;;;; warnings.

(defvar *connection-failure-counter* 0)
(defvar *connection-failure-limit* 23)

(defun lep-login (fd password)
  (sb-sys:enable-interrupt sb-unix:sigpipe :ignore)
  (let* ((stream (sb-bsd-sockets:socket-make-stream fd :input t :output t
						    :buffering :line))
	 (ibuf-length (sb-impl::fd-stream-ibuf-length stream)))
    ;; XXX kludge
    (setf (sb-impl::fd-stream-ibuf-length stream) 1)
    (unwind-protect
	(let ((protocol))
	  (handler-case
	      (let* ((protocol-token (read-token stream))
		     (process-name (and protocol-token (read-token stream)))
		     (read-password (and process-name (read-token stream))))
		(unless (and read-password
			     (eql (parse-integer read-password) password))
		  (error "LEP login: incorrect password"))
		(if (member protocol-token
			    '(":lep" ":listener" ":stream")
			    :test 'string=)
		    (setf protocol (read-from-string protocol-token))
		    (error "LEP: invalid protocol: ~A" protocol-token))
		(setf (name (get-thread)) process-name))
	    (error (c)
	      (incf *connection-failure-counter*)
	      (cond
		((< *connection-failure-counter* *connection-failure-limit*)
		  (warn "connection attempt to LEP listener failed: ~A" c))
		((= *connection-failure-counter* *connection-failure-limit*)
		  (warn "~A (lep::*connection-failure-limit* exceeded, ~
                         will ignore further errors)"
			c)))
	      (return-from lep-login)))
	  #+verbose-lep
	  (format t "~&; LEP: Accepted connection for protocol ~S.~%" protocol)
	  (setf (sb-impl::fd-stream-ibuf-length stream) ibuf-length)
	  (ecase protocol
	    (:listener (listener stream))
	    (:lep (lep-loop stream))
	    (:stream (return-lep-stream stream) (setf stream nil))))
      (when stream
	(ignore-errors (close stream))))))

(defun listener (stream)
  (sb-unix::unix-setsid)
  (let* (;; XXX We aren't supposed to touch *terminal-io*.  But.
	 (*terminal-io* stream)
	 (*standard-input* *terminal-io*)
	 (*standard-output* *standard-input*)
	 (*error-output* *standard-input*)
	 (*debug-io* *standard-input*)
	 (*query-io* *standard-input*)
	 (*trace-output* *standard-input*))
    (catch 'exit-listener
      (let ((buffer-number (read stream)))
	(setf (gethash buffer-number *buffer-threads*)
	      (sb-thread:current-thread-id)))
      (sb-thread:with-new-session ()
	(sb-thread::get-foreground)
	(sb-impl::toplevel-repl nil)))))

(defun error-reply (stream session-id str &rest arguments)
  (sb-thread:with-mutex (*send-lock*)
    (write (list session-id :error (apply #'format nil str arguments))
	   :stream stream
	   :case :downcase
	   :readably t)
    (force-output stream)))

(defvar *sessions*)

(defclass session () ((id :initarg :id :accessor id)))

(defgeneric call-session (session command &rest args))

(defun register-session (session)
  (setf (gethash (id session) *sessions*) session))

(defun unregister-session (session)
  (remhash (id session) *sessions*))

(defun lep-loop (stream)
  (setf *lep-stream* stream)
  (let ((emacs (read stream)))
    (format t "~&; LEP connected to ~A~%" emacs))
  (let ((*sessions* (make-hash-table)))
    (catch 'lep-loop
      (loop
	(with-simple-restart
	    (continue-lep-loop "Continue processing LEP commands")
	  (process-one-lep-command stream))))))

(defun process-one-lep-command (stream)
  (handler-case
      (let ((command
	     (handler-case
		 (read stream)
	       (stream-error () (throw 'lep-loop nil)))))
	(destructuring-bind (session-id function &rest args)
	    command
	  (declare (symbol function))
	  (cond
	    (session-id
	      (let ((session (gethash session-id *sessions*)))
		(unless session
		  (user-error nil "no such session: ~D" session-id))
		(process-session-request
		 stream
		 (lambda () (apply #'call-session session function args))
		 session-id)))
	    ((not (fboundp function))
	      (format t "~&; LEP: unsupported command: ~S~%" command)
	      (user-error nil "unsupported command: ~A" function))
	    (t
	      (apply function stream args)))))
    (user-error (c)
      (handler-case
	  (error-reply stream nil "~A" c)
	(stream-error (d)
	  (format t "~&; LEP-LOOP: ~A~%" d)
	  (throw 'lep-loop nil))))))

(defun return-lep-stream (stream)
  ;; c.f. MAKE-EDITOR-LISTENER-STREAM
  (progn ;sb-thread:with-mutex (*stream-protocol-lock*) ;###
    ;; XXX kludge
    (setf (sb-impl::fd-stream-ibuf-length stream) 1)
    (let* ((sid (read stream))
	   (queue (gethash sid *stream-protocol-requests*)))
      (declare (ignore queue))
      (setf *session-stream* (cons sid stream))
      #+(or)				;###
      (sb-thread:condition-notify queue))))

(defun make-session
    (stream command &rest args &key session-id &allow-other-keys)
  (declare (symbol command))
  (unless (or (fboundp command) (find-class command nil))
    (format t "~&; LEP: unsupported session command:~%; ~S~%"
	    (cons command args))
    (error-reply stream session-id "unsupported session: ~A" command)
    (return-from make-session))
  (let* ((class (find-class command nil))
	 (fn (if class
		 (let ((session (make-instance class :id session-id)))
		   (lambda () (apply 'call-session session command args)))
		 (lambda () (apply command args)))))
    (process-session-request stream fn session-id)))

(defun process-session-request (stream fn session-id)
  (handler-case
      (let ((results (multiple-value-list (invoke-with-background-stream fn))))
	(sb-thread:with-mutex (*send-lock*)
	  (let ((*package* (find-package :cl)))
	    (write (list* session-id :reply results) :stream stream :readably t))
	  (force-output stream)))
    (user-error (c)
      (handler-case
	  (error-reply stream session-id "~A" c)
	(stream-error (d)
	  (error "MAKE-SESSION: ~A" d))))))

(defun read-in-package (string &optional package)
  (handler-case
      (let ((*package*
	     (cond
	       ((find-package package))
	       (t *package*))))
	(read-from-string string))
    (reader-error (c)
      (user-error "~A" c))))


;;;; LEP session commands (Emacs-to-Lisp)

(defun arglist-session (&key fspec buffer-package &allow-other-keys)
  (let ((fsymbol (read-in-package fspec buffer-package)))
    (unless (fboundp fsymbol)
      (user-error "No such function: ~A" fsymbol))
    (values (princ-to-string fsymbol)
	    (princ-to-string (sb-introspect:function-arglist fsymbol)))))

(defun describe-something-session (&key fspec buffer-package &allow-other-keys)
  (let ((object (read-in-package fspec buffer-package)))
    (with-output-to-string (s) (describe object s))))

(defun evaluation-request (&key text buffer-package &allow-other-keys)
  (let ((package (cond ((find-package buffer-package)) (t *package*)))
	(eof '#:eof)
	value)
    (with-input-from-string (s text)
      (let ((*package* package))
	(with-compilation-unit ()
	  (for ((form = (read s nil eof))
		:until (eq form eof))
	    (setf value (eval form))))))
    (format nil "~S" value)))

(defun eval-from-emacs-session (&key string buffer-package &allow-other-keys)
  (let ((*package* (cond ((find-package buffer-package)) (t *package*))))
    (format nil "~S" (eval (read-from-string string)))))

(defun tcp-simulate-special-char (fn n)
  (funcall fn (gethash n *buffer-threads*)))

(defun interrupt-tcp-listener (pid)
  (sb-unix:unix-kill pid sb-unix:sigint))

(defun apropos-session (&key string &allow-other-keys)
  (let ((*package* (find-package :cl)))
    (with-output-to-string (*standard-output*)
      (apropos string))))

(defun starts-with-p (string prefix)
  (let ((mismatch (mismatch string prefix)))
    (or (null mismatch) (= mismatch (length prefix)))))

(defun list-all-completions-session
    (&key pattern functions-only-p &allow-other-keys)
  (setf pattern (string-upcase pattern))
  (with-collector (collect)
    (dolist (package (list-all-packages))
      (loop for symbol being each present-symbol in package do
	    (when (and (eq (symbol-package symbol) package)
		       (starts-with-p (string-upcase (symbol-name symbol))
				      pattern)
		       (or (not functions-only-p)
			   (fboundp symbol)))
	      (collect symbol))))))

(defun toggle-trace (&key fspec buffer-package &allow-other-keys)
  (let ((fsymbol (read-in-package fspec buffer-package)))
    (cond
      ((member fsymbol (trace))
	(eval `(untrace ,fsymbol))
	(values fsymbol t))
      (t
	(eval `(trace ,fsymbol))
	(values fsymbol nil)))))

;; C-c C		fi:lisp-who-calls
;; C-c c		fi:list-who-calls
;; C-c k		fi:kill-definition

(defun function-documentation-session
    (&key fspec buffer-package &allow-other-keys)
  (let ((fsymbol (read-in-package fspec buffer-package)))
    (cond
      ((documentation fsymbol 'function))
      (t "NIL"))))

(defun macroexpand-session (&key form package expander &allow-other-keys)
  (let ((parsed-form (read-in-package form package)))
    (princ-to-string (funcall (the symbol expander) parsed-form))))

(defclass scm::metadot-session (session)
    ((argument :accessor argument)
     (definitions :accessor definitions)))

(defmethod call-session ((session scm::metadot-session) command &rest args)
  (if (eq command :request)
      (pop-metadot session)
      (destructuring-bind (&key fspec buffer-package &allow-other-keys)
	  args
	(let ((fsymbol (read-in-package fspec buffer-package))
	      (objects '()))
	  (setf (argument session) fsymbol)
	  (when (find-class fsymbol nil)
	    (push (find-class fsymbol nil) objects))
	  (when (fboundp fsymbol)
	    (let ((fn (cond
			((macro-function fsymbol))
			((fdefinition fsymbol)))))
	      (push fn objects)
	      (when (typep fn 'generic-function)
		(dolist (method (sb-mop:generic-function-methods fn))
		  (push method objects)))))
	  (setf (definitions session)
		(remove nil (mapcar #'sb-introspect:find-definition-source
				    (reverse objects)))) 
	  (register-session session)
	  (pop-metadot session)))))

(defun pop-metadot (session)
  (let ((definition (pop (definitions session))))
    (unless (definitions session)
      (unregister-session session))
    (unless definition
      (user-error "No definitions for ~A" (argument session)))
    (unless (sb-introspect:definition-source-pathname definition)
      (user-error "Sorry, source file for ~A not recorded" (argument session)))
    (values
     (namestring
      (truename
       (sb-introspect:definition-source-pathname definition)))
     (sb-introspect:definition-source-character-offset definition)
     (length (definitions session)))))

(defun terminate-session (stream session-id)
  (declare (ignore stream))
  (remhash session-id *sessions*)
  (values))


;;;; Lisp-to-Emacs Requests

(defvar *session-id* 0)

(defun eval-in-emacs (string)
  (ensure-lep)
  (format *lep-stream* "(nil :request lep::eval-from-lisp ~S)" string)
  (force-output *lep-stream*))

(defun send-command (fn &rest args)
  (ensure-lep)
  (sb-thread:with-mutex (*send-lock*)
    (let ((*print-case* :downcase)
	  (*package* (find-package :cl)))
      (write `(nil :request lep::make-session-for-lisp ,*session-id* nil nil
		   ,fn ,@args)
	     :stream *lep-stream*
	     :readably t))
    (force-output *lep-stream*)
    (n++ *session-id*)))

(defun make-editor-listener-stream (&rest keys)
  ;; keys seem to be: PARENT X Y WIDTH HEIGHT SPLITP NAME NO-SELECT
  (let (stream)
    (progn ;;sb-thread:with-mutex (*stream-protocol-lock*) ;###
      (setf *session-stream* nil)
      (let ((sid (send-command 'lep::create-listener-stream keys))
	    (queue (sb-thread:make-waitqueue)))
	(setf (gethash sid *stream-protocol-requests*) queue)
	#+(or)				;###
	(sb-thread:condition-wait queue *stream-protocol-lock*)
	(sb-ext:with-timeout 20
	  (let ((w 1e-3))
	    (loop
	      (let ((s *session-stream*))
		(when s
		  (assert (eql (car s) sid))
		  (setf stream (cdr s))
		  (return)))
	      (format *trace-output* "~&; waiting...~%")
	      (force-output *trace-output*)
	      (sleep w)
	      (setf w (* w 2)))))))
    (let ((buffer-number (read stream)))
      (setf (gethash buffer-number *buffer-threads*)
	    (sb-thread:current-thread-id)))
    stream))


;;;; background streams

(defun make-background-stream (&rest options)
  (make-instance 'background-stream :options options))

(defun invoke-with-background-stream (function &rest options)
  (with-open-stream
      ;; XXX We aren't supposed to touch *terminal-io*.  But.
      (*terminal-io* (apply #'make-background-stream options))
    (let* ((*standard-output* *terminal-io*)
	   ;; *trace-output* should stay unchanged
	   (*standard-input* *standard-output*)
	   (*query-io* *standard-output*)
	   (*debug-io* *standard-output*)
	   (*error-output* *standard-output*))
      ;; rebind *session-lock*, else sb-thread will think we still talk to
      ;; sb-sys::*tty* and block us before entering the debugger
      (sb-thread:with-new-session ()
	(sb-thread::get-foreground)
	(funcall function)))))

(defmacro with-background-stream ((&rest options) &body body)
  `(invoke-with-background-stream (lambda () ,@body) ,@options))

(sb-simple-streams:def-stream-class background-stream
    (sb-simple-streams:dual-channel-simple-stream)
  ((options :initarg :options :accessor options)
   (fdstream :initform nil :accessor %fdstream))
  (:default-initargs :options '()))

(defun maybe-open-buffer (bgstream)
  (unless (%fdstream bgstream)
    (let* ((fdstream (apply #'make-editor-listener-stream (options bgstream)))
	   (fd (sb-sys:fd-stream-fd fdstream)))
      (setf (%fdstream bgstream) fdstream)
      (setf (sb-simple-streams:stream-input-handle bgstream) fd)
      (setf (sb-simple-streams:stream-output-handle bgstream) fd))))

(defmethod sb-simple-streams:device-open ((stream background-stream) options)
  (declare (ignore options))
  (sb-simple-streams:with-stream-class (background-stream stream)
    (sb-simple-streams:add-stream-instance-flags
     stream :simple :input :output :dual)

    ;; Kann mich nicht erinnern in ACL Laengenslots gesetzt zu haben.
    (let ((length (sb-simple-streams:device-buffer-length stream)))
      (setf (sb-simple-streams::sm sb-simple-streams::buffer stream)
	    (make-string length))
      (setf (sb-simple-streams::sm sb-simple-streams::out-buffer stream)
	    (make-string length))
      (setf (sb-simple-streams::sm sb-simple-streams::buf-len stream) length)
      (setf (sb-simple-streams::sm sb-simple-streams::max-out-pos stream)
	    length))

    ;; sb-simple-streams::*std-control-out-table* gibt's in ACL auch nicht
    (setf (sb-simple-streams::sm sb-simple-streams::control-out stream)
	  sb-simple-streams::*std-control-out-table*)
    
    (setf (stream-external-format stream) :default))
  t)

(defmethod sb-simple-streams:device-read
    ((stream background-stream) buffer start end blocking)
  (maybe-open-buffer stream)
  (call-next-method))

(defmethod sb-simple-streams:device-write
    ((stream background-stream) buffer start end blocking)
  (when (or (and (null buffer) (not (eql start end)))
	    ;; XXX SBCL bug -- :flush is not even a legal argument!
	    (eq buffer :flush))
    (sb-simple-streams::with-stream-class (background-stream stream)
      ;; XXX SBCL bug -- END must not be NIL
      (setf end (sb-simple-streams::sm sb-simple-streams::outpos stream)))
    (when (= start end)
      (return-from sb-simple-streams:device-write)))
  (maybe-open-buffer stream)
  (call-next-method))

(defmethod sb-simple-streams:device-close ((bgstream background-stream) abort)
  (declare (ignore abort))
  (awhen (%fdstream bgstream) (close %))
  t)

(defun background-thread-p (pid)
  (let ((session sb-thread::*session*))
    (sb-thread:with-mutex ((sb-thread::session-lock session))
      (not (eql (car (sb-thread::session-interactive-threads session)) pid)))))

(defun resolve-synonym-stream (s)
  (while (typep s 'synonym-stream)
    (setf s (symbol-value (synonym-stream-symbol s))))
  s)

(defun maybe-divert-debugger (condition invoke-debugger-hook)
  (declare (ignore invoke-debugger-hook))
  (when (and (lep-is-running)
	     (background-thread-p (sb-thread:current-thread-id))
	     (eq (resolve-synonym-stream *debug-io*) sb-sys::*tty*))
    (with-open-stream (*debug-io* (make-background-stream))
      ;; let the debugger appear in an emacs buffer and tell sb-thread that
      ;; we will not interfere with the terminal
      (sb-thread:with-new-session ()
	(sb-thread::get-foreground)
	(invoke-debugger condition)))))

(setf sb-debug::*invoke-debugger-hook* 'maybe-divert-debugger)
